// Implementation for Configuration
//
// Copyright Matthias Gaertner, Schenefeld, Germany, 2001
// You may use this code under the terms of the GPL.
// Updated packages may be found at http://www.rtner.de/parsled/
//

#include <Configuration.h>
#include <ConfigLine.h>

//---

_ConfigurationListElement::_ConfigurationListElement()
{
    _next = NULL;
    _c = NULL;
    _name = NULL;
}

_ConfigurationListElement::~_ConfigurationListElement()
{
    delete _c;
    delete _name;
}

//---

ConfigSourceFile::ConfigSourceFile( FILE *f )
{
    _f = f;
}

bool ConfigSourceFile::eof( void ) const
{
    return (feof(_f) != 0);
}

int ConfigSourceFile::nextc( void )
{
    return fgetc(_f);
}

//---

ConfigSourceString::ConfigSourceString( const SimpleString *s )
{
    _s = s;
    _n = 0;
}

bool ConfigSourceString::eof( void ) const
{
    return (_n < _s->length());
}

int ConfigSourceString::nextc( void )
{
    return _s->getAt(_n++);
}

//---

Configuration::Configuration() :
    _l( NULL ),
    _cursor( NULL ),
    _cursor2( NULL )
{
}

Configuration::~Configuration()
{
    reset();
}

void Configuration::reset( void )
{
    _reset( _l );
    _l = NULL;
}

void Configuration::_reset( struct _ConfigurationListElement *p )
{
    if( p->_next != NULL )
    {
        _reset( p->_next );
        p->_next = NULL;
    }
    delete p;
}

int Configuration::feedConfiguration( IConfigurationSource *ics )
{
    SimpleString currentSection("",0);
    IConfigClass *currentClass = createClassForSection( &currentSection );
    addSection( &currentSection, currentClass );

    while( !ics->eof() )
    {
        IConfigLine *line = createLine( &currentSection );
        line->feed( ics );

        if( line->isComment() )
        {
            // do nothing
        }
        else if( line->isSection() )
        {
            notifySectionStop( &currentSection, currentClass );
            currentSection.assign( line->getToken() );
            if( (currentClass=getSection(&currentSection)) == NULL )
            {
                currentClass = createClassForSection( &currentSection );
                if( currentClass == NULL )
                {
                    notifySectionStop( &currentSection, currentClass );
                    return 401;
                }
                addSection( &currentSection, currentClass );
            }
            notifySectionStart( &currentSection );
        }
        else
        {
            int r;
            if( (r=currentClass->feedLine( line )) != 0 )
            {
                notifySectionStop( &currentSection, currentClass );
                return r;
            }
            notifyLine( line );
        }
        delete line;
    }
    notifySectionStop( &currentSection, currentClass );
    return 0;
}

int Configuration::feedConfiguration( const SimpleString *s )
{
    ConfigSourceString css( s );
    return feedConfiguration( &css );
}

int Configuration::feedConfiguration( FILE *f )
{
    ConfigSourceFile csf( f );
    return feedConfiguration( &csf );
}

IConfigClass *Configuration::getSection( SimpleString *s ) const
{
    const struct _ConfigurationListElement *p = _l;
    while( p != NULL )
    {
        if( s->equals( p->_name ) )
        {
            return p->_c;
        }
        p = p->_next;
    }
    return NULL;
}

void Configuration::resetIter( void )
{
    _cursor = _l;
}

void Configuration::pushIter( void )
{
    _cursor2 = _cursor;
    _cursor = _l;
}

void Configuration::popIter( void )
{
    _cursor = _cursor2;
}

IConfigClass *Configuration::getNextIter( void )
{
    if( _cursor == NULL )
    {
        return NULL;
    }
    IConfigClass *r = _cursor->_c;
    _cursor = _cursor->_next;
    return r;
}

//---

IConfigLine *Configuration::createLine( const SimpleString *currentSection )
{
    return new ConfigLine();
}

void Configuration::addSection( const SimpleString *section, IConfigClass *cl )
{
    struct _ConfigurationListElement **pp = &_l;
    while( *pp != NULL )
    {
        pp = &(*pp)->_next;
    }

    *pp = new struct _ConfigurationListElement;
    (*pp)->_c = cl;
    (*pp)->_name = &section->clone();
}

struct _ConfigurationListElement *Configuration::unlinkSectionByName( const SimpleString *section )
{
    struct _ConfigurationListElement **pp = &_l;
    while( *pp != NULL )
    {
        if( section->equals( (*pp)->_name ) )
        {
            struct _ConfigurationListElement *p = *pp;
            *pp = (*pp)->_next;
            return p;
        }
        pp = &(*pp)->_next;
    }
    return NULL;
}

struct _ConfigurationListElement *Configuration::unlinkSectionByClass( const IConfigClass *cl )
{
    struct _ConfigurationListElement **pp = &_l;
    while( *pp != NULL )
    {
        if( cl == (*pp)->_c )
        {
            struct _ConfigurationListElement *p = *pp;
            *pp = (*pp)->_next;
            return p;
        }
        pp = &(*pp)->_next;
    }
    return NULL;
}

/* */
